#include "../tuning/CURRENT_GPU"
#include "float.h"
#include "aspen_devel.h"


#if !defined(GPU_ARCH)
#define	GPU_ARCH	(-1)
#endif


#if GPU_ARCH == Tesla
#define	USE_VOLATILE	1
#define	USE_TEXTURE	1
#define	USE_RESTRICT	0
#define	USE_LDG		0
#else
#if GPU_ARCH == Fermi
#define	USE_VOLATILE	1
#define	USE_TEXTURE	1
#define	USE_RESTRICT	0
#define	USE_LDG		0
#else
#if GPU_ARCH == Kepler
#define	USE_VOLATILE	1
#define	USE_TEXTURE	0
#define	USE_RESTRICT	1
#define	USE_LDG		0
#else
#if GPU_ARCH == Kepler2
#define	USE_VOLATILE	1
#define	USE_TEXTURE	0
#define	USE_RESTRICT	0
#define	USE_LDG		1
#else
#if GPU_ARCH == Maxwell
#define	USE_VOLATILE	1
#define	USE_TEXTURE	0
#define	USE_RESTRICT	0
#define	USE_LDG		1
#else
#if GPU_ARCH == Maxwell2
#define	USE_VOLATILE	1
#define	USE_TEXTURE	0
#define	USE_RESTRICT	0
#define	USE_LDG		1
#else
error !!!
#endif // Maxwell2
#endif // Maxwell
#endif // Kepler2
#endif // Kepler
#endif // Fermi
#endif // Tesla

#include "aspen_devel_postdef.h"

#if USE_TEXTURE
DEF_TEXTURE(texture_X, X)
#define	__TRICK_offsetX	,offsetX
#else
#define	__TRICK_offsetX
#endif

#define	SUMUP_PREFIX		_NAME_1PARAMS_(SGEMVT,GPU_PREFIX)
#define	do_best_kernel_SGEMVT	_NAME_1PARAMS_(SUMUP_PREFIX,Do_Best_Kernel)
#define	SGEMV_t_STUB		_NAME_1PARAMS_(SGEMV_t,GPU_PREFIX)


#if CURRENT_GPU==GPU_ARCH
#include "../template/sumup_1Dx2_template.incl"
#include "../template/gemv_t_template.incl"
#endif


#if CURRENT_GPU==GPU_ARCH
#if 1
#include "../tuning/sgemvt-at-opt.c"
#else
long
do_best_kernel_SGEMVT ( int blk,
			const scalar_t alpha, const scalar_t beta,
			const scalar_t * a, const scalar_t * x,
			scalar_t * y,
			int lda,
			int n1, int n2, int offsetX,
			double * ff )
{
  GEMVT_KERNEL < GPU_ARCH, scalar_t, 32, 1, 1, 1 >
    ( alpha, beta, a, x, y, lda, n1, n2, offsetX );
  return (long)0;
}
#endif
#endif


extern "C" __host__ void
API(SGEMV_t_STUB)(int blk,
		  const scalar_t alpha, const scalar_t beta,
		  const scalar_t * a, const scalar_t * x,
		  scalar_t * y,
		  const int lda,
		  const int n1, const int n2)
{
  // y = alpha * A^t * x + beta * y
  // A(1:n2,1:n1) :: A^T(1:n1,1:n2)
  // y(1:n1)
  // x(1:n2)

#if CURRENT_GPU==GPU_ARCH

  if ( n1 <= 0 || n2 <= 0 ) return;
  if ( alpha == CONST(0) && beta == CONST(1) ) return;

  size_t offsetX = 0;

#if USE_TEXTURE
  SETUP_TEXTURE ( texture_X, x, n2, offsetX );
#endif

  double f;
  do_best_kernel_SGEMVT ( blk, alpha, beta, a, x, y, lda, n1, n2,
                          offsetX, &f );
#if USE_TEXTURE
  cudaUnbindTexture ( texture_X );
#endif

#else
  fprintf( stderr,
           "Caution, ASPEN is tuned up for %s.\n",
           ASPEN_GPU_Names(CURRENT_GPU) );
#endif

}

