#if GPU_ARCH>=Kepler

template < class TYPE >
__forceinline__ __device__ TYPE
__SHFL_xor ( TYPE const x, int const laneMask, int const width )
{
  return  __shfl_xor( x, laneMask, width );
}

template < >
__forceinline__ __device__ double
__SHFL_xor < double > ( double const x, int const laneMask, int const width )
{
  int tmp_hi, tmp_lo;
#if 1
  asm volatile ( "mov.b64  {%1, %0}, %2;\n\t"
                 :"=r"(tmp_hi), "=r"(tmp_lo)
                 :"d"(x) );
#else
  tmp_hi = __double2hiint( x );
  tmp_lo = __double2loint( x );
#endif
  tmp_hi = __shfl_xor( tmp_hi, laneMask, width );
  tmp_lo = __shfl_xor( tmp_lo, laneMask, width );
  return  __hiloint2double( tmp_hi, tmp_lo );
}


#define	SwapElements(x,y,flag, ... )                                    \
  _MACRO_WRAP_(                                                         \
               TYPE  xx = x;                                            \
               x = flag ? y : xx;                                       \
               y = flag ? xx : y;                                       \
                                                                        )

#define	_sumup_inter_Warp( ... )					\
  _MACRO_WRAP_(                                                         \
               w += local_id;						\
               if ( BLOCK_SIZE == 128 ) {				\
                 if ( c >= 2 ) { w[0] = x0; }                           \
                 __syncthreads ( );                                     \
                 if ( c < 2 ) {                                         \
                   x0 += w[64];                                         \
                   if ( c ) { w[0] = x0; }                              \
                 }                                                      \
                 __syncthreads ( );                                     \
                 if ( c == 0 ) { x0 += w[32]; }                         \
               } else if ( BLOCK_SIZE > 32 ) {				\
                 if ( c ) { w[0] = x0; }				\
                 __syncthreads ( );                                     \
                 if ( c == 0 ) {					\
                   TYPE  * W = w + 32;                                  \
                   for ( int i=32; i<BLOCK_SIZE; i+=32 ) {              \
                     x0 += (* W); W += 32;                              \
                   }                                                    \
                 }                                                      \
               }							\
                                                                        )

#define	ADD_Default( x, lane, ... )                                     \
  _MACRO_WRAP_(                                                         \
               x += __SHFL_xor < TYPE > ( x,  lane, 32 );               \
                                                                        )

#define	ADD_WithSwap( x, y, lane, ... )                                 \
  _MACRO_WRAP_(                                                         \
               int flag = local_id & lane;                              \
               SwapElements( x, y, flag );                              \
               x += __SHFL_xor < TYPE > ( y,  lane, 32 );               \
                                                                        )

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce_in_block ( TYPE x0, TYPE * w, TYPE * z )
{

  int  local_id = threadIdx.x;
  int  c = local_id / 32;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    for ( int I_ = 1; I_ < 32; I_ *= 2 ) {
      ADD_Default( x0, I_ );
    }
  }

  if ( local_id == 0 ) z[0] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce2_in_block ( TYPE x0, TYPE x1, TYPE * w, TYPE * z )
{

  int  local_id = threadIdx.x;
  int  c = local_id / 32;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    for ( int I_ = 2; I_ < 32; I_ <<= 1 ) {
      ADD_Default( x0, I_ );
    }
  }

  if ( local_id < 2 ) z[local_id] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce3_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE * w, TYPE * z )
{

  int local_id = threadIdx.x;
  int  c = local_id / 32;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
    ADD_Default( x2,  1 );
    ADD_WithSwap( x0, x2, 2 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    for ( int I_ = 4; I_ < 32; I_ <<= 1 ) {
      ADD_Default( x0, I_ );
    }
  }

  if ( local_id < 3 ) z[local_id] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce4_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE * w, TYPE * z )
{

  int local_id = threadIdx.x;
  int  c = local_id / 32;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
    ADD_WithSwap( x2, x3, 1 );
    ADD_WithSwap( x0, x2, 2 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    for ( int I_ = 4; I_ < 32; I_ <<= 1 ) {
      ADD_Default( x0, I_ );
    }
  }

  if ( local_id < 4 ) z[local_id] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce5_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE * w, TYPE * z )
{

  int local_id = threadIdx.x;
  int  c = local_id / 32;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
    ADD_WithSwap( x2, x3, 1 );
    ADD_WithSwap( x0, x2, 2 );
    ADD_Default( x4,  1 );
    ADD_Default( x4,  2 );
    ADD_WithSwap( x0, x4, 4 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    for ( int I_ = 8; I_ < 32; I_ <<= 1 ) {
      ADD_Default( x0, I_ );
    }
  }

  if ( local_id < 5 ) z[local_id] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce6_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE x5, TYPE * w, TYPE * z )
{

  int local_id = threadIdx.x;
  int  c = local_id / 32;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
    ADD_WithSwap( x2, x3, 1 );
    ADD_WithSwap( x0, x2, 2 );
    ADD_WithSwap( x4, x5, 1 );
    ADD_Default( x4, 2 );
    ADD_WithSwap( x0, x4, 4 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    for ( int I_ = 8; I_ < 32; I_ <<= 1 ) {
      ADD_Default( x0, I_ );
    }
  }

  // __syncthreads();
  if ( local_id < 6 ) z[local_id] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce7_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE x5, TYPE x6, TYPE * w, TYPE * z )
{

  int local_id = threadIdx.x;
  int  c = local_id / 32;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
    ADD_WithSwap( x2, x3, 1 );
    ADD_WithSwap( x0, x2, 2 );
    ADD_WithSwap( x4, x5, 1 );
    ADD_Default( x6, 1 );
    ADD_WithSwap( x4, x6, 2 );
    ADD_WithSwap( x0, x4, 4 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    for ( int I_ = 8; I_ < 32; I_ <<= 1 ) {
      ADD_Default( x0, I_ );
    }
  }

  // __syncthreads();
  if ( local_id < 7 ) z[local_id] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce8_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE x5, TYPE x6, TYPE x7, TYPE * w, TYPE * z )
{

  int local_id = threadIdx.x;
  int  c = local_id / 32;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
    ADD_WithSwap( x2, x3, 1 );
    ADD_WithSwap( x0, x2, 2 );
    ADD_WithSwap( x4, x5, 1 );
    ADD_WithSwap( x6, x7, 1 );
    ADD_WithSwap( x4, x6, 2 );
    ADD_WithSwap( x0, x4, 4 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    for ( int I_ = 8; I_ < 32; I_ <<= 1 ) {
      ADD_Default( x0, I_ );
    }
  }

  // __syncthreads();
  if ( local_id < 8 ) z[local_id] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce9_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE x5, TYPE x6, TYPE x7, TYPE x8, TYPE * w, TYPE * z )
{

  int local_id = threadIdx.x;
  int  c = local_id / 32;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
    ADD_WithSwap( x2, x3, 1 );
    ADD_WithSwap( x0, x2, 2 );
    ADD_WithSwap( x4, x5, 1 );
    ADD_WithSwap( x6, x7, 1 );
    ADD_WithSwap( x4, x6, 2 );
    ADD_WithSwap( x0, x4, 4 );
    ADD_Default( x8,  1 );
    ADD_Default( x8,  2 );
    ADD_Default( x8,  4 );
    ADD_WithSwap( x0, x8, 8 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    for ( int I_ = 16; I_ < 32; I_ <<= 1 ) {
      ADD_Default( x0, I_ );
    }
  }

  // __syncthreads();
  if ( local_id < 9 ) z[local_id] = x0;
}

#endif

