#!/bin/sh

CG=$1
KERNEL_KEY=$2

CG1=$3.cu
CG2=$4.cu
CG3=$5.cu
CG3_5=$6.cu
CG5=$7.cu
CG5_2=$8.cu


COMPILER=$CUDA_PATH'/bin/nvcc -O3 --ptxas-options=-v --maxrregcount=63 -DAT_VERSION --fmad=true -I'$CUDA_PATH'/include -I./include -I../include -I../../include -I. '


touch comp-log.$$; \rm comp-log.$$

if [ $CG -eq 130 ]; then
	$COMPILER -DGPU_ARCH=Tesla \
		-arch=compute_13 -code=compute_13,sm_13 \
		-c $CG1 -o tmp$$.o >& comp-log.$$
fi

if [ $CG -eq 200 ]; then
	$COMPILER -DGPU_ARCH=Fermi \
		-arch=compute_20 -code=compute_20,sm_20 \
		-c $CG2 -o tmp$$.o >& comp-log.$$
fi

if [ $CG -eq 300 ]; then
	$COMPILER -DGPU_ARCH=Kepler \
		-arch=compute_30 -code=compute_30,sm_30 \
		-c $CG3 -o tmp$$.o >& comp-log.$$
fi

if [ $CG -eq 350 ]; then
	$COMPILER -DGPU_ARCH=Kepler2 \
		-arch=compute_35 -code=compute_35,sm_35 \
		-c $CG3_5 -o tmp$$.o >& comp-log.$$
fi

if [ $CG -eq 500 ]; then
	$COMPILER -DGPU_ARCH=Maxwell \
		-arch=compute_50 -code=compute_50,sm_50 \
		-c $CG5 -o tmp$$.o >& comp-log.$$
fi

if [ $CG -eq 520 ]; then
	$COMPILER -DGPU_ARCH=Maxwell2 \
		-arch=compute_52 -code=compute_52,sm_52 \
		-c $CG5_2 -o tmp$$.o >& comp-log.$$
fi

cat comp-log.$$ | \
	awk '/Func/ { \
		if ( $0 ~ /'$KERNEL_KEY'/ ) { \
			Name = $7; \
			gsub(/.*'$KERNEL_KEY'/, "", Name); \
			gsub(/_.*/, "", Name); \
			gsub(/[A-Za-z]/, " ", Name); \
			split(Name, P); \
			spilled=0; \
			while ( 1 ) { \
				getline; \
				if ( $0~/spill/ ) { \
					spilled=$5+$9; \
				} \
				if ( $0~/registers/ ) { \
					Reg = $5; \
					if ( spilled > 192 ) { \
						s = 0; \
					} else { \
						s = (spilled==0)?1:2; \
					} \
					print Reg" "P[2]" "P[3]" "P[4]" "P[5]" "P[6]" "s; \
					break; \
				} \
			} \
		} \
	}'


\rm comp-log.$$
\rm tmp$$.o

exit 0

